/* Atomix -- a little mind game about atoms and molecules.
 * Copyright (C) 1999 Jens Finke
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <parser.h>
#include <tree.h>

#include "global.h"
#include "main.h"
#include "theme.h"
#include "callbacks.h"
#include "support.h"
#include "goal.h"

#define SCALE_FACTOR  0.6

void
goal_print_offset(gpointer ptr, gpointer data);

GnomeCanvasItem*
create_small_moveable(gdouble x, gdouble y, AtomixObject* obj);

Goal*
goal_new(PlayField* pf)
{
    Goal* goal;
    gint row,col;
    AtomixObject* obj;
    GSList* list;
    guint id, current_id = 0;
    ItemOffset* offset;
    gint max_id = 0; 

    goal = g_malloc(sizeof(Goal));    
    
    goal->pf = pf;
    goal->canvas_items = NULL;
    
    /* initialise index */
    goal->index = g_ptr_array_new();
    g_assert(goal->index);
    
    while(current_id <= max_id)
    {
	list = NULL;
	
	for(row = 0; row < pf->n_rows; row++)
	{
	    for(col = 0; col < pf->n_cols; col++)
	    {
		obj = playfield_get_object(pf, row, col);
		if(obj->type == ATOMIX_MOVEABLE)
		{
		    id = obj->id;
		    
		    if(id == current_id)
		    {
			offset = g_malloc(sizeof(ItemOffset));
			
			offset->horiz = col;
			offset->vert = row;
		    
			list = g_slist_append(list, offset);
		    }

		    max_id = MAX(max_id, id);
		    
		}
	    }
	}
	g_ptr_array_add(goal->index, (gpointer)list);
	    
	current_id++;
    }

    return goal;
}

void
goal_destroy(Goal* goal)
{
    g_ptr_array_free(goal->index, TRUE);
    g_free(goal);
}

void 
goal_print(Goal* goal)
{
    guint i;
    GSList* list;
    
    if(goal->pf)
    {
	g_print("GOAL:\n");
	playfield_print(goal->pf);
    }
#ifdef DEBUG
    else 
    {
	g_print("No goal specified.\n");
    }
#endif

    if(goal->index)
    {
	for(i = 0; i < goal->index->len; i++)
	{
	    list = (GSList*)g_ptr_array_index(goal->index,i);
	    g_print("%x: ",i);
	    if(list)
	    {
		g_slist_foreach(list, (GFunc)goal_print_offset, NULL);
	    }
	    else
	    {
		g_print("list NULL.");
	    }
	    g_print("\n");
	}
    }
#ifdef DEBUG
    else
    {
	g_print("Index is somehow NULL!\n");
    }
#endif
}

void
goal_print_offset(gpointer ptr, gpointer data)
{
    ItemOffset* off = (ItemOffset*) ptr;
    g_print("{%x|%x} ",off->horiz, off->vert);
}

gboolean
goal_reached(Goal* goal, PlayField* pf, guint row_anchor, guint col_anchor)
{
    guint id;
    gint i;
    gint r,c;
    gint end_row, end_col;
    gint start_row, start_col;
    GSList* list;
    AtomixObject* goal_obj;
    AtomixObject* pf_obj;
    gint goal_row, goal_col;
    ItemOffset* offset;
    gboolean result = FALSE;
    gboolean item_checked = FALSE;
    gboolean item_result = TRUE;

    id = playfield_get_object(pf, row_anchor, col_anchor)->id;
    list = (GSList*) g_ptr_array_index(goal->index, id);
    
    for(i = 0; i < g_slist_length(list) && !result; i++)
    {
	offset = (ItemOffset*) g_slist_nth_data(list, i);

	start_row = row_anchor - offset->vert;
	start_col = col_anchor - offset->horiz;

	item_checked = FALSE;
	item_result = TRUE;

	if((start_row >= 0) && (start_col >= 0))
	{
	    item_checked = TRUE;

	    goal_row = goal_col = 0;
	    end_row = start_row + goal->pf->n_rows;
	    end_col = start_col + goal->pf->n_cols;

	    for(r = start_row; (r < end_row) && item_result; r++)
	    {
		goal_col = 0;
		for(c = start_col; (c < end_col) && item_result; c++)
		{
		    goal_obj = playfield_get_object(goal->pf,goal_row,goal_col);
		    pf_obj = playfield_get_object(pf,r,c);

		    if(goal_obj->type == ATOMIX_MOVEABLE)
		    {
			
			if((goal_obj->id != pf_obj->id) ||
			   (goal_obj->type != pf_obj->type))
			{
			    item_result = FALSE;
			}		    
		    }
		    goal_col++;
		}
		goal_row++;
	    }

	    result = item_checked && item_result;	    
	}
    }

    return result;
}

GnomeCanvasItem*
create_small_moveable(gdouble x, gdouble y, AtomixObject* obj)
{
    GnomeCanvas *canvas;
    GdkImlibImage *img = NULL;
    GnomeCanvasItem *item = NULL;
    Theme* theme;
    gchar* file;
    
    theme = get_actual_theme();
    file = theme_get_object_image(theme, ATOMIX_MOVEABLE, obj->id);
    
    img = gdk_imlib_load_image(file);
    
    if(img) {
	canvas = GNOME_CANVAS(lookup_widget(get_app(), "ca_goal"));
	item = gnome_canvas_item_new(gnome_canvas_root(canvas),
				     gnome_canvas_image_get_type(),
				     "image", img,
				     "x", x,
				     "y", y,
				     "width", (double)(img->rgb_width)*SCALE_FACTOR,
				     "height", (double)(img->rgb_height)*SCALE_FACTOR,
				     "anchor", GTK_ANCHOR_NW,
				     NULL);                              
	gtk_signal_connect (GTK_OBJECT (item), "destroy",
			    (GtkSignalFunc) free_imlib_image, img);  
	
	
    } 
#ifdef DEBUG
    else {
	g_print("Atom image not found!\n");
    }
#endif    
    
    return item;
}

void
goal_render(Goal* goal)
{
    GnomeCanvasItem *item;
    GnomeCanvas *canvas;
    gint row,col;
    gdouble x;
    gdouble y;
    AtomixObject* obj;
    AtomixType type;
    gint tile_width, tile_height;

    canvas = GNOME_CANVAS(lookup_widget(get_app(), "ca_goal"));
    item = gnome_canvas_item_new(gnome_canvas_root(canvas),
				 gnome_canvas_rect_get_type(),
				 "x1", 0.0,
				 "y1", 0.0,
				 "x2", 150.0,
				 "y2", 150.0,
				 "outline_color", "black",
				 "fill_color", "SkyBlue",
				 "width_pixels", 1,
				 NULL);    
    for(row=0; row < goal->pf->n_rows; row++) 
    {
	for(col=0; col < goal->pf->n_cols; col++)
	{
	    obj = playfield_get_object(goal->pf, row, col);
	    type = obj->type;
	    switch(type)
	    {
	    case ATOMIX_MOVEABLE:
		theme_tile_size(get_actual_theme(), &tile_width, &tile_height);

		x = col * tile_width * SCALE_FACTOR;
		y = row * tile_height * SCALE_FACTOR;
		item = create_small_moveable(x, y, obj);
		goal->canvas_items = g_slist_append(goal->canvas_items, (gpointer) item);
		break;
		
	    case ATOMIX_OBSTACLE:
	    case ATOMIX_NONE:
	    default:
#ifdef DEBUG
		g_print("goal render: ignore other object type ...\n");
#endif
	    }
	}
    }
}

void
goal_clear(Goal* goal)
{
    g_slist_foreach(goal->canvas_items, destroy_item, NULL);
    g_slist_free(goal->canvas_items);
    goal->canvas_items = NULL;    
}

