/* Atomix -- a little mind game about atoms and molecules.
 * Copyright (C) 1999 Jens Finke
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "playfield.h"

PlayField*
playfield_new(void)
{
    PlayField* pf;
    int i,j;

    pf = g_malloc(sizeof(PlayField));

    if(pf) 
    {
	/* initialise matrix */
	pf->n_rows = 0;
	pf->n_cols = 0;

	pf->selector_row = -1;
	pf->selector_col = -1;

	for(i=0; i < MAX_ROWS; i++) {
	    for(j=0; j < MAX_COLS; j++) {
		pf->matrix[i][j] = NULL;
	    }
	}
    }
#ifdef DEBUG
    else {
	g_print("Not enough memory for Matrix allocation!");
    }
#endif

    return pf;
}
gboolean
playfield_add_row(PlayField* pf)
{
    gboolean successful = FALSE;
    int c;

    if(pf->n_rows+1 <= MAX_ROWS)
    {
	if(pf->n_cols > 0) 
	{
	    /* add appropriate numbers of cols to this row */
	    for(c=0; c < pf->n_cols; c++)
	    {
		pf->matrix[pf->n_rows][c] = atomix_object_new();
	    }
	}

	pf->n_rows++;
	successful = TRUE;
    }

    return successful;
}

gboolean
playfield_add_column(PlayField* pf)
{
    gboolean successful = FALSE;
    int r;
    
    if(pf->n_cols+1 <= MAX_COLS)
    {
	if(pf->n_rows > 0)
	{
	    /* add in every existing row a column */
	    for(r=0; r < pf->n_rows; r++)
	    {
		pf->matrix[r][pf->n_cols] = atomix_object_new();
	    }
	}
	pf->n_cols++;
	successful = TRUE;
    }
    
    return successful;
}

void
playfield_create_border(PlayField* pf) 
{
    int i;

    if(pf!=NULL) {
	
	for(i=0; i < pf->n_rows; i++) 
	{
	    pf->matrix[i][0]->type = ATOMIX_OBSTACLE;
	    pf->matrix[i][pf->n_cols-1]->type = ATOMIX_OBSTACLE;
	}
	
	for(i=1; i < pf->n_cols; i++) 
	{
	    pf->matrix[0][i]->type = ATOMIX_OBSTACLE;
	    pf->matrix[pf->n_rows-1][i]->type = ATOMIX_OBSTACLE;
	}
    }
}

void
playfield_destroy(PlayField* pf)
{
    int i,j;

    for(i = 0; i < MAX_ROWS; i++) 
    {
	for(j = 0; j < MAX_COLS; j++)
	{
	    if(pf->matrix[i][j])
	    {
		g_free(pf->matrix[i][j]);
	    }
	}
    }
    g_free(pf);
}

PlayField*
playfield_copy(PlayField* pf)
{
    PlayField* pf_copy = NULL;
    guint row,col;

    if(pf) 
    {
	pf_copy = playfield_new();
	g_assert(pf_copy!=NULL);
	for(row=0; row < pf->n_rows; row++) 
	{
	    playfield_add_row(pf_copy);
	    for(col=0; col < pf->n_cols; col++)
	    {
		if(row==0)
		{
		    playfield_add_column(pf_copy);
		}
		playfield_set_object(pf_copy, atomix_object_copy(playfield_get_object(pf,row,col)), row, col );

	    }
	}    
    }
#ifdef DEBUG
    else
    {
	g_print("pf equals NULL.\n");
    }
#endif
    
    return pf_copy;
}

void
playfield_swap_tiles(PlayField* pf, guint src_row, guint src_col,
		     guint dest_row, guint dest_col)
{
    AtomixObject* obj;
    obj = playfield_get_object(pf,src_row,src_col);
    pf->matrix[src_row][src_col] = playfield_get_object(pf,dest_row, dest_col);
    pf->matrix[dest_row][dest_col] = obj;
}


AtomixObject*
playfield_clear_tile(PlayField* pf, guint row, guint col)
{
    AtomixObject* obj;

    if(pf!=NULL && 
       (row < pf->n_rows) && (row >= 0 )&&
       (col < pf->n_cols) && (col >= 0)) 
    {	
	obj = playfield_get_object(pf, row, col);
	playfield_set_object(pf, atomix_object_new(), row,col);
	
	return obj;
    }
#ifdef DEBUG
    else {
	g_print("playfield.c: Array out of bounds or pointer equals NULL!");
    }
#endif    
    
    return NULL;
}

void
playfield_set_type(PlayField* pf, AtomixType type, guint row, guint col)
{
    if(pf!=NULL && 
       (row < pf->n_rows) && (row >= 0 )&&
       (col < pf->n_cols) && (col >= 0)) 
    {
	pf->matrix[row][col]->type = type;
    }
#ifdef DEBUG
    else {
	g_print("playfield.c: Array out of bounds or pointer equals NULL!");
    }
#endif
}

AtomixType
playfield_get_type(PlayField *pf, guint row, guint col)
{
    if(pf!=NULL && 
       (row < pf->n_rows) && (row >= 0 )&&
       (col < pf->n_cols) && (col >= 0)) 
    {
	return pf->matrix[row][col]->type;
    }
#ifdef DEBUG
    else {
	g_print("playfield.c: Array out of bounds or pointer equals NULL!");
    }
#endif
    
    return 0;
}

AtomixObject*
playfield_get_object(PlayField *pf, guint row, guint col)
{
    if(pf!=NULL && 
       (row < pf->n_rows) && (row >= 0 )&&
       (col < pf->n_cols) && (col >= 0)) 
    {
	return pf->matrix[row][col];
    }

#ifdef DEBUG
    else {
	g_print("playfield.c: Array out of bounds or pointer equals NULL!");
    }
#endif
    
    return NULL;
}

void
playfield_set_object(PlayField* pf, AtomixObject* obj, guint row, guint col)
{
    if(pf!=NULL && 
       (row < pf->n_rows) && (row >= 0 )&&
       (col < pf->n_cols) && (col >= 0)) 
    {
	if(pf->matrix[row][col])
	{
	    atomix_object_destroy(pf->matrix[row][col]);
	}
	pf->matrix[row][col] = obj;
    }

#ifdef DEBUG
    else {
	g_print("playfield.c: Array out of bounds or pointer equals NULL!");
    }
#endif
}

void
playfield_get_selector(PlayField* pf, gint* row, gint* col)
{
    if(pf!=NULL) 
    {
	*row = pf->selector_row;
	*col = pf->selector_col;
    }
}

void
playfield_print(PlayField* pf)
{
    int i, j;
    AtomixObject* obj;

    if(pf)
    {
	g_print("N_ROWS: %x\n",pf->n_rows);
	g_print("N_COLS: %x\n",pf->n_cols);
	
	for(i=0; i < pf->n_rows; i++) 
	{
	    for(j=0; j < pf->n_cols; j++)
	    {
		obj = playfield_get_object(pf,i,j); 
		if(obj)
		{
		    g_print("%d|%d ",
			    obj->type,
			    obj->id);
		}
		else
		{
		    g_print("NULL ");
		}
	    }
	    g_print("\n");
	}
	g_print("\n");
    }
    else
    {
	g_print("Cant print playfield because of NULL pointer.\n");
    }
}


PlayField* 
playfield_load_xml(xmlNodePtr pf_node)
{
    xmlNodePtr row_node;
    xmlNodePtr tile_node;
    PlayField *pf;
    gchar *prop_value;
    gint row, col;

    pf = playfield_new();
    

    /* reading number of columns and rows 
       (currently the rows and cols will be ignored, but
       for later use here is the functionality)
    prop_value = xmlGetProp(pf_node, "rows");
    pf->n_rows = atoi(prop_value);
    prop_value = xmlGetProp(pf_node, "cols");
    pf->n_cols = atoi(prop_value);
    */   

    /* read the matrix */
    row = col = 0;
    
    row_node = pf_node->childs;
    while(row_node != NULL)
    {
	if(g_strcasecmp(row_node->name, "ROW")==0)
	{
	    playfield_add_row(pf);
	    tile_node = row_node->childs;
	    col = 0;
	    
	    while(tile_node != NULL)
	    {	    
		if(g_strcasecmp(tile_node->name, "TILE")==0)
		{
		    AtomixObject *obj;
		    AtomixType type;
		    gint id;
		    
		    prop_value = xmlGetProp(tile_node, "type");
		    type = atoi(prop_value);
		    
		    prop_value = xmlGetProp(tile_node, "imgid");
		    id = atoi(prop_value);
		    
		    if(row==0)
		    {
			playfield_add_column(pf);
		    }
		    
		    obj = atomix_object_new_args(type, id);
		    playfield_set_object(pf, obj, row, col);
		    col++;
		}
		else
		{
		    g_print("Unexpected Tag (<%s>), ignoring.\n", tile_node->name);
		}
		
		tile_node = tile_node->next;
	    }
	    row++;
	}
	else
	{
	    g_print("Unexpected Tag (<%s>), ignoring.\n", row_node->name);
	}

	row_node = row_node->next;
    }
    
    return pf;
}

/*
 * Creates the XML representation of an PlayField.
 * pf = Playfield to save
 * pf_node = reference to the <PLAYFIELD> Tag node
 */
void playfield_save_xml(PlayField *pf, xmlNodePtr pf_node)
{
    gint row, col;
    xmlAttrPtr attr;
    xmlNodePtr row_node;
    xmlNodePtr tile_node;
    gchar *str_buffer; 
    gint length;
    AtomixType type;
    gint imgid;

    str_buffer = g_malloc(5*sizeof(gchar));
    
    /* number of columns and rows */
    length = g_snprintf(str_buffer, 5, "%i", pf->n_rows);
    attr = xmlSetProp(pf_node, "rows", g_strdup(str_buffer));
    length = g_snprintf(str_buffer, 5, "%i", pf->n_cols);
    attr = xmlSetProp(pf_node, "cols", g_strdup(str_buffer));

    /* add every tile */
    for(row = 0; row < pf->n_rows; row++)
    {	
	row_node = xmlNewChild(pf_node, NULL, "ROW", NULL);

	for(col = 0; col < pf->n_cols; col++)
	{
	    tile_node = xmlNewChild(row_node, NULL, "TILE", NULL);

	    /* add tile type */
	    type = playfield_get_object(pf, row, col)->type;
	    length = g_snprintf(str_buffer, 5, "%i", (gint)type);
	    attr = xmlSetProp(tile_node, "type", str_buffer);

	    /* add image id */
	    imgid = playfield_get_object(pf, row, col)->id;
	    length = g_snprintf(str_buffer, 5, "%i", imgid);
	    attr = xmlSetProp(tile_node, "imgid", str_buffer);		
	}
    }
    
    g_free(str_buffer);		         
}
